#include <stdio.h>
#include <strings.h>
#include <ctype.h>

#define N 5000
#define M 500
#define L 50
#define DEBUG 0

char codes[M][L], words[N][L];
int seen[M], refs[N], pos[N];
static int n, m;

int letter_kw(c)
char c;
{
  if (c >= 'a' && c <= 'z' || c >= '0' && c <= '9' || c == '*' || c == '_') return 1;
  else return 0;
}

int letter_txt(c)
char c;
{
  if (c >= 'a' && c <= 'z' || c >= '0' && c <= '9') return 1;
  else return 0;
}

int open_kw(filename)
char filename[];
{
  int i, j, c, mode;
  FILE *f;

  f = fopen(filename,"r");
  if (f == NULL) {
    fprintf(stderr,"No such file: %s\n",filename);
    return 0;
  }
  for (i=0; i < N; i++) words[i][0] = '_';
  mode = n = m = i = 0;

  while ((c = getc(f)) != EOF) {
    c = tolower(c);

    if (mode == 0) {
      if (letter_kw(c)) codes[m][i++] = c;
      else if (c == ':' || c == ';') {
        codes[m][i] = '\0';
        strcpy(&words[n][1],codes[m]);
        words[n][1+i] = '_'; words[n][2+i] = '\0'; 
        refs[n] = m;
        if (DEBUG) printf("%s %s %d\n",codes[m],words[n],refs[n]);
        n++; i = 0;
        if (c == ':') mode = 1;
        else m++;
      }
      else if (c == ',') {
        words[n][1+i] = '\0';
        fprintf(stderr,"Error: forgotten :,; while reading %s ... %s?\n",codes[m],words[n]);
        return 0;
      }
    }else{
      if (letter_kw(c)) words[n][1+i++] = c;
      else if (c == ',' || c == ';') {
        if (words[n][i] == '*') words[n][i] = '\0';
        else {words[n][1+i] = '_'; words[n][2+i] = '\0';}
        refs[n] = m;
        if (DEBUG) printf("%s %s %d\n",codes[m],words[n],refs[n]);
        n++; i = 0;
        if (c == ';') {mode = 0; m++;}
      }
      else if (c == ':') {
        words[n][1+i] = '\0';
        fprintf(stderr,"Error: forgotten :,; while reading %s ... %s?\n",codes[m],words[n]);
        return 0;
      }
    }
  }
  if (n >= N || m >= M) fprintf(stderr,"Too many keywords (n=%d, m=%d)\n",n,m);
  return 1;
}

int init_kw()
{
  int i;
  for (i=0; i < n; i++) pos[i] = 1;
  /* Give all words leading space */
  return n;
}

int reset_kw()
{
  int i;
  init_kw();
  for (i=0; i < m; i++) seen[i] = 0;
  return m;
}

int check_kw(line)
char line[];
{
  int i,j,stop;
  char c, lastc;
  lastc = '_';
  stop = j = 0;
  init_kw();
  do {
    c = tolower(line[j++]);
    if (c == '\0') {c = '_'; stop = 1;}
    if (!letter_txt(c)) c = '_';
    if (c == '_' && lastc == '_') continue;

    for (i=0; i < n; i++)
    if (!seen[refs[i]]) { 
      if (words[i][pos[i]] != c) {
        pos[i] = 0;
        if (lastc =='_' && words[i][1] == c) pos[i] = 2;
      }
      else if (words[i][++(pos[i])] == '\0') {
        seen[refs[i]] = 1;
        if (DEBUG) fprintf(stderr,"%s:\n%s\n",words[i],line);
      }
    }
    lastc = c;
  } while (!stop);
  return 1;
}

int report_kw(f,verbose)
FILE *f;
int verbose;
{
  int i;
  for (i=0; i < m; i++) {
    if (seen[i]) fprintf(f,"%s ",codes[i]);
    if (seen[i] && verbose) fprintf(stderr,"%s ",codes[i]);
  }
  reset_kw();
  if (verbose) fprintf(stderr,"\n");
  return m;
}

int show_kw()
{
  int i,j;
  for (i=j=0; i < m; i++) {
    printf("\n%s:\t",codes[i]);
    while (refs[j] == i) {
      printf("%s,",words[j]);
      j++;
    }
  }
  printf("\n\n");
}

int grep_kw(argc,argv)
int argc; char *argv[];
{
  int i, j;
  for (i=0; i < argc; i++) {
    words[i][0] = '_';
    for (j=0; argv[i][j] != '\0'; j++)
      words[i][j+1] = tolower(argv[i][j]);
    if (words[i][j] == '*') 
      words[i][j] = '\0';
    else {
      words[i][j+1] = '_';
      words[i][j+2] = '\0';
    }
  }
  m = n = argc;
}

int seen_kw()
{
  int i, yes;
  yes = 0;
  for (i=0; i < m && !yes; i++)
    if (seen[i]) yes = 1;
  reset_kw();
  return yes;
}
